/*============================================================================*/
/* eCall ANSI C fixed-point reference source code                             */
/*                                                                            */
/* File:         ecall_control.c                                              */
/* Version:      8.4.0 (Rel8) / 9.2.0 (Rel9)                                  */
/* Date:         2010-06-15                                                   */
/* Description:  control message functions                                    */
/*----------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>

#include "ecall_control.h"
#include "ecall_defines.h"
#include "ecall_sync.h"
#include "ecall_rom.h"


/*============================================================================*/
/* UTILITY FUNCTION: CtrlSync                                                 */
/*----------------------------------------------------------------------------*/
/* Description: control message sync function                                 */
/*                                                                            */
/* InOut:  CtrlRxData*  control  <-> control struct                           */
/* In:     const Int16* pcm       -> input frame of 16bit PCM samples         */
/*----------------------------------------------------------------------------*/

void CtrlSync(CtrlRxData *control, const Int16 *pcm)
{
  Int16 i;
  CtrlPort *port = &control->port;
  SyncState *sync = control->sync;

  if (sync->events < control->dlSyncLock) {
    sync->flag = False;
    sync->delay = 0;
    sync->offset = 0;

    /* keep control receiver inactive while sync is running */
    port->state = CtrlRxIdle;
    port->u.rx.dlData = DlMsgNoop;
    port->u.rx.dlMetric = 0;

    /* run synchronization */
    Sync(sync, pcm, port->owner, port->invert);

    if (sync->flag == True && sync->delay <= 0) {
      if (sync->index > NRF_DLCHUNK - 1 ||
        sync->amplitude[2] < sync->amplitude[0])
      {
        /* increase number of successful sync trials */
        sync->trials++;
        port->state = CtrlRxSync;

        /* check for muting gaps passed already (avoids unnecessary delays) */
        if (sync->delay < (-PCM_LENGTH)) {
          sync->offset = -(sync->delay)/PCM_LENGTH;
          sync->index -= sync->offset;
          sync->delay += sync->offset*PCM_LENGTH;
        }
        /* count occurrences of the same sync position */
        if (
          sync->events > 0 &&
          sync->delayMem == sync->delay &&
          sync->index%(NRF_DLCHUNK) == 0)
        {
          sync->events++;
        } else {
          sync->events = 1;
        }
        sync->delayMem = sync->delay;
        sync->index = sync->offset;

        if (sync->events >= control->dlSyncLock) {
          control->dlRead = Plus;
          control->dlIndex = sync->offset;

          /* lock sync for control message detection */
          port->state = CtrlRxLock;
          sync->trials = 0;

          LogInfo("[%-4s] sync locked, start control message detection\n----",
            port->owner);

          /* apply sync inversion flag */
          if ((port->invert = sync->invert) == True) {
            LogInfo("[%-4s] codec inversion detected!\n----", port->owner);
          }
        } else {
          sync->amplitude[2] = sync->amplitude[0];
          sync->amplitude[0] = 0;
          sync->npeaks = 0;
        }
      } else {
        sync->amplitude[2] = sync->amplitude[0];
        sync->amplitude[0] = 0;
        sync->npeaks = 0;
      }
    }
  } else {
    /* sync is locked, check whether locked sync is still valid */
    port->state = CtrlRxSync;
    SyncCheck(sync, pcm, port->owner);

    if (sync->index == NRF_DLCHUNK) {
      sync->index = 0;

      if (sync->check == Plus) {
        /* lower-layer control message */
        control->dlRead = Plus;
        sync->checkCnt = 0;
      }
      else if (sync->check == Minus) {
        /* higher-layer control message */
        control->dlRead = Minus;
        sync->checkCnt = 0;
      }
      else {
        /* following message is ignored */
        control->dlRead = Zero;
        sync->checkCnt++;

        for (i = 0; i < 4; i++) {
          sync->corrCheck[i] = 0;
          sync->corrIndex[i] = 0;
        }
        LogInfo("[%-4s] message ignored (bad sync check)", port->owner);
      }
    }
    /* full restart allows IVS to resynchronize if PSAP requests it */
    if (sync->checkCnt >= SYNC_BADCHECKS) {
      port->u.rx.dlData = DlMsgReset;
      port->u.rx.dlMetric = 0;
    } else {
      port->u.rx.dlData = DlMsgNoop;
      port->u.rx.dlMetric = 0;
    }
  }
}

/*============================================================================*/
/* CONTROL FUNCTION: CtrlTxReset                                              */
/*----------------------------------------------------------------------------*/
/* Description: reset function control transmitter                            */
/*                                                                            */
/* InOut:  CtrlTxData* control  <-> control struct                            */
/* In:     const char* owner     -> modem identification                      */
/*----------------------------------------------------------------------------*/

void CtrlTxReset(CtrlTxData *control, const char *owner)
{
  control->port.state  = CtrlTxIdle;
  control->port.invert = False;
  control->port.owner  = owner;

  /* initial port data */
  control->port.u.tx.dlData = DlMsgNoop;
  control->port.u.tx.dlIndex = -1;
}

/*============================================================================*/
/* CONTROL FUNCTION: CtrlTxProcess                                            */
/*----------------------------------------------------------------------------*/
/* Description: process function control transmitter                          */
/*                                                                            */
/* InOut:  CtrlTxData* control  <-> control struct                            */
/*         Int16* pcm           <-> frame of 16bit PCM samples                */
/*----------------------------------------------------------------------------*/

void CtrlTxProcess(CtrlTxData *control, Int16 *pcm)
{
  CtrlPort *port = &control->port;

  /* PSAP transmitter, initialize PCM data to zero */
  memset(pcm, 0, sizeof(Int16)*PCM_LENGTH);

  /* safety check on index range */
  if (port->u.tx.dlIndex >= 0 && port->u.tx.dlIndex < NRF_DLCHUNK) {
    if (port->u.tx.dlData != DlMsgNoop) {
      CtrlTxMod(pcm, port->u.tx.dlData, port->u.tx.dlIndex);
    }
  } else if (port->u.tx.dlIndex >= NRF_DLCHUNK) {
    Abort("message index too large (value: %d), exit.", port->u.tx.dlIndex);
  }
}

/*============================================================================*/
/* UTILITY FUNCTION: CtrlTxMod                                                */
/*----------------------------------------------------------------------------*/
/* Description: control message transmitter using prestored sequences         */
/*                                                                            */
/* In:     Int16  symbol  -> lower-layer or higher-layer message symbol       */
/*         Int16  index   -> position within message frame                    */
/* Out:    Int16* pcm     <- output data                                      */
/*----------------------------------------------------------------------------*/

void CtrlTxMod(Int16 *pcm, Int16 symbol, Int16 index)
{
  Int16 i;
  Int16 lpcm = sizeof(Int16)*PCM_LENGTH;

  if (!(symbol >= 0 && symbol < 4) &&
    !(symbol >= SET_LLMSG && symbol < SET_LLMSG + SET_HLMSG))
  {
    Abort("message symbol out of range, exit.");
  }
  if (index < NRF_WAKEUP) {
    /* wakeup tone */
    for (i = 0; i < PCM_LENGTH; i++)
      pcm[i] = wakeupSin500[i%16];
  }
  else if (index < NRF_SYNC) {
    /* sync frame */
    memcpy(pcm, &syncFrame[(index-NRF_WAKEUP)*PCM_LENGTH], lpcm);

    /* invert sync preamble for higher-layer messages */
    if (symbol >= SET_LLMSG) {
      for (i = 0; i < PCM_LENGTH; i++)
        pcm[i] = -pcm[i];
    }
    /* add signal offset */
    for (i = 0; i < PCM_LENGTH; i++) {
      pcm[i] += (pcm[i] == 0) ? 12000 : 5000;
    }

    /* wakeup tone extends into the first frame of the preamble */
    if (NRF_WAKEUP > 0 && index == NRF_WAKEUP) {
      for (i = 0; i < 32; i++)
        pcm[i] = wakeupSin500[i%16];
    }
  }
  else {
    i = (symbol < SET_LLMSG)
      ? index - NRF_DLMUTE1LL - NRF_SYNC
      : index - NRF_DLMUTE1HL - NRF_SYNC;

    if (symbol < SET_LLMSG && i >= 0 && i < NRF_DLDATA) {
      /* data message lower-layer symbol */
      memcpy(pcm, &dlPcmData[symbol][i*PCM_LENGTH], lpcm);
    }
    else if (symbol >= SET_LLMSG && i >= 0 && i < NRF_DLDATA) {
      /* 1st part of data message higher-layer symbol */
      memcpy(pcm, &dlPcmData[(symbol-SET_LLMSG)%4][i*PCM_LENGTH], lpcm);
    }
    else if (symbol >= SET_LLMSG && i >= NRF_DLDATA && i < 2*NRF_DLDATA) {
      /* 2nd part of data message higher-layer symbol */
      i -= NRF_DLDATA;
      memcpy(pcm, &dlPcmData[((symbol-SET_LLMSG)/4)%4][i*PCM_LENGTH], lpcm);
    }
    else {
      /* muting frame */
      memset(pcm, 0, lpcm);
    }
  }
}

/*============================================================================*/
/* CONTROL FUNCTION: CtrlRxReset                                              */
/*----------------------------------------------------------------------------*/
/* Description: reset function control receiver                               */
/*                                                                            */
/* InOut:  CtrlRxData* control   <-> control struct                           */
/* In:     const char* owner      -> modem identification                     */
/*         SyncState*  sync       -> pointer to sync struct                   */
/*         Int16*      buffer     -> pointer to control receiver buffer       */
/*         Int16       syncLock   -> number of sync events required           */
/*----------------------------------------------------------------------------*/

void CtrlRxReset(CtrlRxData *control, const char *owner,
                 SyncState *sync, Int16 *buffer, Int16 syncLock)
{
  control->port.state  = CtrlRxIdle;
  control->port.invert = False;
  control->port.owner  = owner;

  /* initial port data */
  control->port.u.rx.dlData = DlMsgNoop;
  control->port.u.rx.dlMetric = 0;

  /* set control pointers */
  control->sync = sync;
  control->buffer = buffer;

  /* set internal variables */
  control->dlHlackData = 0;
  control->dlRead = Zero;
  control->dlIndex = 0;
  control->dlSyncLock = syncLock;

  /* clear memory */
  memset(control->buffer, 0, sizeof(Int16)*(NRF_MEMCTRL*PCM_LENGTH));
}

/*============================================================================*/
/* CONTROL FUNCTION: CtrlRxProcess                                            */
/*----------------------------------------------------------------------------*/
/* Description: process function control receiver                             */
/*                                                                            */
/* InOut:  CtrlRxData*  control  <-> control struct                           */
/* In:     const Int16* pcm       -> input frame of 16bit PCM samples         */
/*----------------------------------------------------------------------------*/

void CtrlRxProcess(CtrlRxData *control, const Int16 *pcm)
{
  Int16 i;
  Int16 metric;
  CtrlPort *port = &control->port;
  SyncState *sync = control->sync;

  /* increase sync index */
  sync->index++;

  if (control->dlRead == Zero) {
    /* run synchronization (sync function to set port data) */
    CtrlSync(control, pcm);
  }
  else if (control->dlRead == Plus &&
    control->dlIndex == NRF_DLMUTE1LL + NRF_DLDATA - 1)
  {
    /* enough lower-layer data received, detect data now */
    i = (NRF_MEMCTRL - NRF_DLDATA)*PCM_LENGTH;
    port->u.rx.dlData =
      CtrlRxDemod(&control->buffer[i + sync->delayMem], &metric);
    port->u.rx.dlMetric = metric;

    /* prepare for next sync */
    control->dlRead = Zero;
    control->dlIndex = 0;

    /* reinitialize sync check parameters */
    for (i = 0; i < 4; i++) {
      sync->corrCheck[i] = 0;
      sync->corrIndex[i] = 0;
    }
    sync->check = Zero;
  }
  else if (control->dlRead == Minus &&
    control->dlIndex == NRF_DLMUTE1HL + 2*NRF_DLDATA - 1)
  {
    /* enough higher-layer data received, detect first data portion */
    i = (NRF_MEMCTRL - 2*NRF_DLDATA)*PCM_LENGTH;
    port->u.rx.dlData =
      CtrlRxDemod(&control->buffer[i + sync->delayMem], &metric);
    port->u.rx.dlMetric = metric;

    /* detect second data portion and combine */
    i = (NRF_MEMCTRL - NRF_DLDATA)*PCM_LENGTH;
    port->u.rx.dlData +=
      4*CtrlRxDemod(&control->buffer[i + sync->delayMem], &metric);
    port->u.rx.dlData += SET_LLMSG;
    port->u.rx.dlMetric = MIN(port->u.rx.dlMetric, metric);

    /* prepare for next sync */
    control->dlRead = Zero;
    control->dlIndex = 0;

    /* reinitialize sync check parameters */
    for (i = 0; i < 4; i++) {
      sync->corrCheck[i] = 0;
      sync->corrIndex[i] = 0;
    }
    sync->check = Zero;
  }
  else {
    /* sync detected, wait for enough data to process */
    control->dlIndex++;
    port->u.rx.dlData = DlMsgNoop;
    port->u.rx.dlMetric = 0;
  }
}

/*============================================================================*/
/* UTILITY FUNCTION: CtrlRxDemod                                              */
/*----------------------------------------------------------------------------*/
/* Description: control message receiver                                      */
/*                                                                            */
/* In:     const Int16* pcm     -> input PCM buffer                           */
/* Out:    Int16*       metric  <- reliability factor (-1: skip)              */
/* Return: Int16                <- demodulated message                        */
/*----------------------------------------------------------------------------*/

Int16 CtrlRxDemod(const Int16 *pcm, Int16 *metric)
{
  Int16 k, n, msg;
  Int32 xCorr[4];
  Int32 maxCorr, runnerup;

  /* correlation receiver */
  for (k = 0; k < 4; k++) {
    xCorr[k] = 0;
    for (n = 0; n < NRF_DLDATA*PCM_LENGTH; n++)
      xCorr[k] += (pcm[n]*dlPcmDataMatch[k][n]) >> 12;
  }

  /* find largest correlation value */
  maxCorr = xCorr[0];
  msg = 0;
  for (k = 1; k < 4; k++) {
    if (xCorr[k] > maxCorr) {
      maxCorr = xCorr[k];
      msg = k;
    }
  }
  /* find second largest correlation value */
  runnerup = xCorr[(msg+1)%4];
  for (k = 0; k < 4; k++) {
    if (xCorr[k] > runnerup && xCorr[k] < maxCorr)
      runnerup = xCorr[k];
  }
  /* test on reliability by absolute and relative value */
  if (maxCorr < IVS_THRESHOLD) {
    *metric = -1;
  }
  else if ((*metric = (Int16)((maxCorr-runnerup)/(maxCorr/100))) <= 33) {
    *metric = -1;
  }
  return msg;
}
