/*============================================================================*/
/* eCall ANSI C fixed-point reference source code                             */
/*                                                                            */
/* File:         ecall_modem.c                                                */
/* Version:      8.4.0 (Rel8) / 9.2.0 (Rel9)                                  */
/* Date:         2010-06-15                                                   */
/* Description:  modulator and demodulator functions                          */
/*----------------------------------------------------------------------------*/

#include <string.h>

#include "modemx.h"
#include "ecall_modem.h"
#include "ecall_defines.h"
#include "ecall_fec.h"
#include "ecall_rom.h"


/*============================================================================*/
/* IVS FUNCTION: IvsTransmitter                                               */
/*----------------------------------------------------------------------------*/
/* Description: IVS transmitter function                                      */
/*                                                                            */
/* In:     const ModState* ms      -> modulator struct                        */
/*         const Ord1*     buffer  -> code bit buffer                         */
/*         Int16           rv      -> redundancy version                      */
/*         Int16           index   -> position within uplink frame            */
/* Out:    Int16*          pcm     <- output data                             */
/*----------------------------------------------------------------------------*/

void IvsTransmitter(const ModState *ms, const Ord1 *buffer, Int16 *pcm,
                    Int16 rv, Int16 index)
{
  Int16 i, j, k, mgPos, symIdx;
  Int16 redIdx = ((rv%2) ? 0 : 1)*(ARQ_MAX/2) + (rv/2)%(ARQ_MAX/2);

  /* treat first redundancy version differently */
  if (rv == 0) {
    if (index < NRF_WAKEUP) {
      /* output wakeup tone */
      mgPos = -1;
      for (i = 0; i < PCM_LENGTH; i++) {
        pcm[i] = ms->wakeupSin[i%(ms->wutperiod)];
      }
    } else if (index < NRF_SYNC) {
      /* output sync frame portion */
      mgPos = -1;
      memcpy(pcm, &syncFrame[PCM_LENGTH*(index-NRF_WAKEUP)],
        sizeof(Int16)*PCM_LENGTH);

      /* wakeup tone extends into the first frame of the preamble */
      if (NRF_WAKEUP > 0 && index == NRF_WAKEUP) {
        for (i = 0; i < 32; i++)
          pcm[i] = ms->wakeupSin[i%(ms->wutperiod)];
      }
    } else {
      mgPos = index - NRF_SYNC;
    }
  } else {
    /* no sync frames in subsequent redundancy versions */
    mgPos = index;
  }
  if (mgPos >= 0) {
    /* no longer sync, but check for muting gaps */
    if (ms->mgTable[mgPos] >= 0) {

      /* convert code bits to symbol indices and modulate */
      for (i = 0; i < ms->mfpf; i++) {
        k = (ms->mgTable[mgPos]*(ms->mfpf)+i)*(ms->bpsym);

        for (j = 0, symIdx = 0; j < ms->bpsym; j++) {
          symIdx += buffer[redVerIndex[redIdx][k+j]] << ((ms->bpsym)-1-j);
        }
        SymbolMod(ms, symIdx, &pcm[i*(ms->spmf)]);
      }
    } else {
      memset(pcm, 0, sizeof(Int16)*PCM_LENGTH);
    }
  }
}

/*============================================================================*/
/* IVS FUNCTION: SymbolMod                                                    */
/*----------------------------------------------------------------------------*/
/* Description: symbol modulator                                              */
/*                                                                            */
/* In:     const ModState* ms      -> modulator struct                        */
/*         Int16           symbol  -> symbol index                            */
/* Out:    Int16*          mPulse  <- modulated output sequence               */
/*----------------------------------------------------------------------------*/

void SymbolMod(const ModState *ms, Int16 symbol, Int16 *mPulse)
{
  Int16 i, shifter;
  Int16 nsym  = 1 << (ms->bpsym);    /* number of symbols */
  Int16 nsym2 = 1 << (ms->bpsym-1);  /* half of the number of symbols */

  if (symbol < 0 || symbol > nsym-1) {
    Abort("symbol index out of range, exit.");
  }
  /* initialize memory */
  for (i = 0; i < ms->spmf; i++) {
    mPulse[i] = 0;
  }
  if (symbol < nsym2) {
    /* creates positive symbols */
    shifter = symbol*(ms->spmf)/nsym2;
    for (i = 0; i < ms->spmf; i++)
      mPulse[(i+shifter)%(ms->spmf)] = ms->ulPulse[i];
  }
  else {
    /* creates negative symbols */
    shifter = (nsym-1-symbol)*(ms->spmf)/nsym2;
    for (i = 0; i < ms->spmf; i++)
      mPulse[(i+shifter)%(ms->spmf)] = (-1)*(ms->ulPulse[i]);
  }
}

/*============================================================================*/
/* IVS FUNCTION: Byte2Bit                                                     */
/*----------------------------------------------------------------------------*/
/* Description: conversion byte vector to bit vector                          */
/*                                                                            */
/* In:     Ord8* in      -> vector of input bytes                             */
/*         Int16 length  -> length of input                                   */
/* Out:    Ord1* out     <- vector of output bits                             */
/*----------------------------------------------------------------------------*/

void Byte2Bit(const Ord8 *in, Ord1 *out, Int16 length)
{
  Int16 i, j;

  for (i = length-1; i >= 0; i--) {
    for (j = 0; j < 8; j++)
      out[j+i*8] = (Ord1)(0x1 & (in[i]>>(7-j)));
  }
}

/*============================================================================*/
/* PSAP FUNCTION: PsapReceiver                                                */
/*----------------------------------------------------------------------------*/
/* Description: PSAP receiver function (decoding is done outside)             */
/*                                                                            */
/* In:     const ModState* ms        -> modulator struct                      */
/*         const Int16*    pcm       -> input data for demodulation           */
/* Out:    IntLLR*         softBits  <- demodulated soft bit sequence         */
/*----------------------------------------------------------------------------*/

void PsapReceiver(const ModState *ms, const Int16 *pcm, IntLLR *softBits)
{
  Int16 m;
  for (m = 0; m < ms->mfpf; m++) {
    SymbolDemod(ms, &pcm[m*(ms->spmf)], &softBits[m*(ms->bpsym)]);
  }
}

/*============================================================================*/
/* PSAP FUNCTION: SymbolDemod                                                 */
/*----------------------------------------------------------------------------*/
/* Description: symbol demodulator                                            */
/*                                                                            */
/* In:     const ModState* ms        -> modulator struct                      */
/*         const Int16*    mPulse    -> received pulse train                  */
/* Out:    IntLLR*         softBits  <- demodulated soft bit sequence         */
/*----------------------------------------------------------------------------*/

void SymbolDemod(const ModState *ms, const Int16 *mPulse, IntLLR *softBits)
{
  Int16 i, j, k, m;
  Int16 nsym  = 1 << (ms->bpsym);    /* number of symbols */
  Int16 nsym2 = 1 << (ms->bpsym-1);  /* half of the number of symbols */

  Int32 Lacc, r;
  Int32 lp[3]; Ord1 bp[3] = {0, 0, 0};
  Int32 lm[3]; Ord1 bm[3] = {0, 0, 0};

  for (i = 0; i < nsym2; i++) {
    for (j = 0, r = 0; j < ms->spmf; j++) {
      r += ((Int32)mPulse[j])*((Int32)(ms->ulPulseMatch[i*(ms->spmf)+j]));
    }
    /* consider positive and negative copy */
    for (j = 0; j <= 1; j++) {
      m = (j) ? (nsym-1-i)*(ms->bpsym) : i*(ms->bpsym);
      r = (j) ? -r : r;

      Lacc = MpyLacc(r, FEC_VAR) - FEC_MEAN;
      Lacc = Lacc >> (21-LOGEXP_QIN);

      if (ABS(Lacc) > LLR_MAX) {
        Lacc = SIGN(Lacc)*LLR_MAX;
      }
      /* cumulative sum of logexp values */
      for (k = 0; k < ms->bpsym; k++)
        if (indexBits[m+k]) {
          if (!bp[k]) {
            lp[k] = Lacc;
            bp[k] = 1;
          } else
            lp[k] = (Int32)JacLog(lp[k], Lacc);
        } else {
          if (!bm[k]) {
            lm[k] = Lacc;
            bm[k] = 1;
          } else
            lm[k] = (Int32)JacLog(lm[k], Lacc);
        }
    }
  }
  /* determine final LLR values */
  for (k = 0; k < ms->bpsym; k++) {
    Lacc = lm[k] - lp[k];

    if (ABS(Lacc) > LLR_MAX) {
      softBits[k] = (IntLLR)(SIGN(Lacc)*LLR_MAX);
    } else {
      softBits[k] = (IntLLR)Lacc;
    }
  }
}

/*============================================================================*/
/* PSAP FUNCTION: Bit2Byte                                                    */
/*----------------------------------------------------------------------------*/
/* Description: conversion bit vector to byte vector                          */
/*                                                                            */
/* In:     const Ord1* in      -> vector of input bits                        */
/*         Int16       length  -> length of output                            */
/* Out:    Ord8*       out     <- vector of output bytes                      */
/*----------------------------------------------------------------------------*/

void Bit2Byte(const Ord1 *in, Ord8 *out, Int16 length)
{
  Int16 i, j;
  memset(out, 0, sizeof(Ord8)*length);

  for (i = 0; i < length; i++) {
    for (j = 0; j < 8; j++)
      out[i] += (Ord8)(in[i*8+j] << (7-j));
  }
}

/*============================================================================*/
/* PSAP FUNCTION: MpyLacc                                                     */
/*----------------------------------------------------------------------------*/
/* Description: multiply 32bit number with 16bit number (32bit result)        */
/*                                                                            */
/* In:     Int32 var32  -> 32bit number                                       */
/*         Int16 var16  -> 16bit number                                       */
/* Return: Int32        <- result                                             */
/*----------------------------------------------------------------------------*/

Int32 MpyLacc(Int32 var32, Int16 var16)
{
  Int16 temp;
  Int32 result;

  temp = (Int16)var32;
  temp = temp >> 1;
  temp = (Int16)0x7fff & (Int16)temp;

  result = (Int32)var16 * (Int32)temp;
  result = result >> 15;
  result = result + (Int32)var16*(var32 >> 16);

  return result;
}

/*============================================================================*/
/* UTILITY FUNCTION: SetModState                                              */
/*----------------------------------------------------------------------------*/
/* Description: set the modulator state                                       */
/*                                                                            */
/* In:     Int16     modType   -> type of modulator to use                    */
/* InOut:  ModState* ms       <-> modulator struct                            */
/*----------------------------------------------------------------------------*/

void SetModState(ModState *ms, ModType modType)
{
  switch (modType) {
    case ModUndef:
      ms->type         = ModUndef;
      ms->bpsym        = 0;
      ms->spmf         = 0;
      ms->mfpf         = 0;
      ms->decpos1      = 0;
      ms->decpos2      = 0;
      ms->wutperiod    = 0;
      ms->nfmute1      = 0;
      ms->nfmute4      = 0;
      ms->nfmuteall    = 0;
      ms->nfdata       = 0;
      ms->ulPulse      = NULL;
      ms->ulPulseMatch = NULL;
      ms->mgTable      = NULL;
      ms->wakeupSin    = NULL;
      ms->wakeupCos    = NULL;
      break;

    case Mod3bit4smp:
      ms->type         = Mod3bit4smp;
      ms->bpsym        = 3;
      ms->spmf         = 16;
      ms->mfpf         = PCM_LENGTH/(ms->spmf);
      ms->decpos1      = 15;
      ms->decpos2      = 30;
      ms->wutperiod    = 16;
      ms->nfmute1      = 1;
      ms->nfmute4      = 3;
      ms->nfmuteall    = 8;
      ms->nfdata       = NRB_CODE_ARQ/((ms->mfpf)*(ms->bpsym));
      ms->ulPulse      = m4smp_ulPulse;
      ms->ulPulseMatch = m4smp_ulPulseMatch;
      ms->mgTable      = m4smp_mgTable;
      ms->wakeupSin    = wakeupSin500;
      ms->wakeupCos    = wakeupCos500;
      break;

    case Mod3bit8smp:
      ms->type         = Mod3bit8smp;
      ms->bpsym        = 3;
      ms->spmf         = 32;
      ms->mfpf         = PCM_LENGTH/(ms->spmf);
      ms->decpos1      = 30;
      ms->decpos2      = 60;
      ms->wutperiod    = 10;
      ms->nfmute1      = 1;
      ms->nfmute4      = 3;
      ms->nfmuteall    = 12;
      ms->nfdata       = NRB_CODE_ARQ/((ms->mfpf)*(ms->bpsym));
      ms->ulPulse      = m8smp_ulPulse;
      ms->ulPulseMatch = m8smp_ulPulseMatch;
      ms->mgTable      = m8smp_mgTable;
      ms->wakeupSin    = wakeupSin800;
      ms->wakeupCos    = wakeupCos800;
      break;

    default:
      Abort("unknown type of modulator, exit.");
      break;
  }
}
