/*============================================================================*/
/* eCall ANSI C fixed-point reference source code                             */
/*                                                                            */
/* File:         ecall_sync.c                                                 */
/* Version:      8.4.0 (Rel8) / 9.2.0 (Rel9)                                  */
/* Date:         2010-06-15                                                   */
/* Description:  synchronization functions                                    */
/*----------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>

#include "ecall_sync.h"
#include "ecall_defines.h"
#include "ecall_rom.h"


/*============================================================================*/
/* UTILITY FUNCTION: SyncReset                                                */
/*----------------------------------------------------------------------------*/
/* InOut:  SyncState* sync       <-> sync struct                              */
/* In:     Int32*     mem         -> pointer to sync memory                   */
/*         Int32*     memWakeup   -> pointer to sync wakeup memory            */
/*----------------------------------------------------------------------------*/

void SyncReset(SyncState *sync, Int32 *mem, Int32 *memWakeup)
{
  /* assign buffer pointers */
  if (mem != NULL) {
    sync->mem = mem;
    sync->memWakeup = memWakeup;
  }
  /* reset variables */
  sync->flag      = False;
  sync->invert    = False;
  sync->delay     = 0;
  sync->delayMem  = 0;
  sync->npeaks    = 0;
  sync->npeaksMem = 0;
  sync->trials    = 0;
  sync->events    = 0;

  /* clear memory */
  sync->amplitude[0] = 0;
  sync->amplitude[1] = 0;
  sync->amplitude[2] = 0;

  /* clear sync subsystem */
  SyncSubReset(&sync->syncPos);
  SyncSubReset(&sync->syncNeg);

  /* initialize control message sync */
  sync->check    = Zero;
  sync->checkCnt = 0;
  sync->index    = NRF_DLMUTE1LL + NRF_DLDATA - 1;
  sync->offset   = 0;

  /* initialize sync check parameters */
  sync->peakPos[0] = NRF_DLCHUNK*PCM_LENGTH - (SYNC_IDXLEN/5-1)*PNSEQ_OSF - 1;
  sync->peakPos[1] = sync->peakPos[0] - PEAK_DIST_PN;
  sync->peakPos[2] = sync->peakPos[0] - PEAK_DIST_PN - PEAK_DIST_PP;
  sync->peakPos[3] = sync->peakPos[0] - PEAK_DIST_NN;

  /* clear sync buffers */
  if (sync->mem != NULL) {
    memset(sync->mem, 0, sizeof(Int32)*NRS_MEMSYNC);
  }
  if (sync->memWakeup != NULL) {
    memset(sync->memWakeup, 0, sizeof(Int32)*2*(NRF_SYNC+1));
  }
}

/*============================================================================*/
/* UTILITY FUNCTION: SyncSubReset                                             */
/*----------------------------------------------------------------------------*/
/* InOut:  SyncSub* ssub  <-> sync subsystem                                  */
/*----------------------------------------------------------------------------*/

void SyncSubReset(SyncSub *ssub)
{
  ssub->flag   = False;
  ssub->delay  = 0;
  ssub->npeaks = 0;
  ssub->trials = 0;

  ssub->amplitude[0] = 0;
  ssub->amplitude[1] = 0;
}

/*============================================================================*/
/* FUNCTION: Sync                                                             */
/*----------------------------------------------------------------------------*/
/* Description: main synchronization function                                 */
/*                                                                            */
/* InOut:  SyncState*   sync    <-> sync struct                               */
/* In:     const Int16* pcm      -> input frame                               */
/*         const char*  caller   -> modem identification                      */
/*         Bool         invert   -> port inversion flag                       */
/*----------------------------------------------------------------------------*/

void Sync(SyncState *sync, const Int16 *pcm, const char *caller, Bool invert)
{
  Int16 i, j, jmax, nframes;
  Int16 idxmax = syncIndexPreamble[SYNC_IDXLEN-1];
  Int32 sgnpcm = (invert == True) ? -1 : 1;

  /* set buffer pointers */
  Int32 *pCorr = &sync->mem[2*idxmax + PCM_LENGTH + 0];
  Int32 *nCorr = &sync->mem[2*idxmax + PCM_LENGTH + 9];
  Int32 *pPos  = &sync->mem[2*idxmax + PCM_LENGTH + 18];
  Int32 *nPos  = &sync->mem[2*idxmax + PCM_LENGTH + 27];

  /* initialize flag */
  sync->flag = False;

  /* test for 500 Hz or 800 Hz sine tone */
  ToneDetect(sync, pcm);

  /* sync filter function */
  jmax = SYNC_IDXLEN-1;
  for (i = -idxmax; i < PCM_LENGTH; i++) {
    while (jmax > 0 && (syncIndexPreamble[jmax]+i) >= PCM_LENGTH) {
      jmax--;
    }
    j = jmax;
    while ((syncIndexPreamble[j]+i) >=0 && j >= 0) {
      sync->mem[i+idxmax] += sgnpcm*
        (Int32)pcm[syncIndexPreamble[j]+i]*
        (Int32)syncPulseForm[j%5]*
        (Int32)syncSequence[j/5];
      j--;
    }
  }
  /* shift all peaks by one slot */
  for (i = 1; i < 9; i++) {
    pCorr[i-1] = pCorr[i];
    nCorr[i-1] = nCorr[i];
    pPos[i-1] = pPos[i] - PCM_LENGTH;
    nPos[i-1] = nPos[i] - PCM_LENGTH;
  }
  pCorr[8] = 0;
  nCorr[8] = 0;

  /* compute maximum value */
  for (i = 0; i < PCM_LENGTH; i++) {
    if (pCorr[8] < sync->mem[i]) {
      pCorr[8] = sync->mem[i];
      pPos[8] = i+1;
    }
    if (nCorr[8] < (-1)*(sync->mem[i])) {
      nCorr[8] = (-1)*(sync->mem[i]);
      nPos[8] = i+1;
    }
  }
  /* shift filter memory by PCM_LENGTH samples */
  nframes = (2*idxmax)/PCM_LENGTH;

  for (i = 1; i <= nframes; i++) {
    memcpy(&sync->mem[(i-1)*PCM_LENGTH], &sync->mem[i*PCM_LENGTH],
      sizeof(Int32)*PCM_LENGTH);
  }
  memcpy(&sync->mem[nframes*PCM_LENGTH], &sync->mem[(nframes+1)*PCM_LENGTH],
    sizeof(Int32)*((2*idxmax)%PCM_LENGTH));

  memset(&sync->mem[nframes*PCM_LENGTH + (2*idxmax)%PCM_LENGTH], 0,
    sizeof(Int32)*PCM_LENGTH);

  /* compute global maximum amplitude */
  if (sync->amplitude != NULL) {
    sync->amplitude[1] = MAX(sync->amplitude[1], MAX(pCorr[8], nCorr[8]));
  }

  /* copy data to subsystem */
  SyncSubPut(sync, &sync->syncPos);
  SyncSubPut(sync, &sync->syncNeg);

  /* peak evaluation for each subsystem */
  SyncSubRun(&sync->syncPos, caller, pPos, pCorr, nPos, nCorr);
  SyncSubRun(&sync->syncNeg, caller, nPos, nCorr, pPos, pCorr);

  /* decision logic */
  if (sync->syncPos.flag) {
    if (sync->syncNeg.flag &&
      sync->syncNeg.amplitude[0] > sync->syncPos.amplitude[0] &&
      sync->syncNeg.npeaks >= sync->syncPos.npeaks)
    {
      sync->invert = True;
      sync->npeaksMem = sync->syncNeg.npeaks;
      SyncSubGet(sync, &sync->syncNeg);
    }
    else if (sync->invert == False || sync->syncPos.npeaks >= sync->npeaksMem)
    {
      sync->invert = False;
      sync->npeaksMem = sync->syncPos.npeaks;
      SyncSubGet(sync, &sync->syncPos);
    }
  } else if (sync->syncNeg.flag &&
    (sync->invert == True || sync->syncNeg.npeaks >= MAX(sync->npeaksMem, 4)))
  {
    sync->invert = True;
    sync->npeaksMem = sync->syncNeg.npeaks;
    SyncSubGet(sync, &sync->syncNeg);
  }
}

/*============================================================================*/
/* IVS FUNCTION: SyncCheck                                                    */
/*----------------------------------------------------------------------------*/
/* Description: check whether locked sync is still valid                      */
/*                                                                            */
/* InOut:  SyncState*   sync    <-> sync struct                               */
/* In:     const Int16* pcm      -> input frame                               */
/*         const char*  caller   -> modem identification                      */
/*----------------------------------------------------------------------------*/

void SyncCheck(SyncState *sync, const Int16 *pcm, const char *caller)
{
  Int16 i, cixMin, cixMax, index, nPeaks[2];

  /* bounds for PCM data as available */
  cixMin = PCM_LENGTH*(sync->index - 1) - sync->delayMem;
  cixMax = PCM_LENGTH*(sync->index - 0) - sync->delayMem;

  for (i = 0; i < 4; i++) {
    while (sync->corrIndex[i] < SYNC_IDXLEN &&
      cixMin <= sync->peakPos[i] + syncIndexPreamble[sync->corrIndex[i]] &&
      cixMax >  sync->peakPos[i] + syncIndexPreamble[sync->corrIndex[i]])
    {
      index = syncIndexPreamble[sync->corrIndex[i]] +
        sync->peakPos[i] + sync->delayMem;

      sync->corrCheck[i] = sync->corrCheck[i] +
        (Int32)pcm[index%PCM_LENGTH]*
        (Int32)syncPulseForm[sync->corrIndex[i]%5]*
        (Int32)syncSequence[sync->corrIndex[i]/5];

      sync->corrIndex[i]++;
    }
  }
  if (sync->index == NRF_DLCHUNK) {
    sync->corrCheck[0] *= (-1);
    sync->corrCheck[3] *= (-1);
    nPeaks[0] = 0;
    nPeaks[1] = 0;

    if (
      sync->corrIndex[0] >= SYNC_IDXLEN-4 &&
      sync->corrIndex[1] == SYNC_IDXLEN &&
      sync->corrIndex[2] == SYNC_IDXLEN &&
      sync->corrIndex[3] == SYNC_IDXLEN)
    {
      for (i = 0; i < 4; i++) {
        if (sync->corrCheck[i] > sync->amplitude[0]/2) {
          nPeaks[0]++;
        }
        else if (-sync->corrCheck[i] > sync->amplitude[0]/2) {
          nPeaks[1]++;
        }
      }
      if (nPeaks[0] > 0) {
        /* this is a lower-layer control message */
        sync->check = Plus;
      }
      else if (nPeaks[1] > 0) {
        /* this is a higher-layer control message */
        sync->check = Minus;
      }
      else {
        /* sync check failure */
        sync->check = Zero;
      }
    } else {
      LogInfo("[%-4s] sync check correlation incomplete!", caller);
    }
  }
}

/*============================================================================*/
/* UTILITY FUNCTION: SyncSubPut, SyncSubGet                                   */
/*----------------------------------------------------------------------------*/
/* InOut:  SyncState* sync  <-> sync struct                                   */
/* InOut:  SyncSub*   ssub  <-> sync subsystem                                */
/*----------------------------------------------------------------------------*/

void SyncSubPut(SyncState *sync, SyncSub *ssub)
{
  ssub->flag         = sync->flag;
  ssub->delay        = sync->delay;
  ssub->npeaks       = sync->npeaks;
  ssub->trials       = sync->trials;
  ssub->amplitude[0] = sync->amplitude[0];
  ssub->amplitude[1] = sync->amplitude[1];
}

void SyncSubGet(SyncState *sync, SyncSub *ssub)
{
  sync->flag         = ssub->flag;
  sync->delay        = ssub->delay;
  sync->npeaks       = ssub->npeaks;
  sync->trials       = ssub->trials;
  sync->amplitude[0] = ssub->amplitude[0];
  sync->amplitude[1] = ssub->amplitude[1];
}

/*============================================================================*/
/* UTILITY FUNCTION: SyncSubRun                                               */
/*----------------------------------------------------------------------------*/
/* Description: sync peak evaluation                                          */
/*                                                                            */
/* InOut:  SyncSub*     ssub    <-> sync subsystem                            */
/* In:     const char*  caller   -> modem identification                      */
/*         const Int32* pPos     -> positive peaks positions                  */
/*         const Int32* pCorr    -> positive peaks correlation values         */
/*         const Int32* nPos     -> negative peaks positions                  */
/*         const Int32* nCorr    -> negative peaks correlation values         */
/*----------------------------------------------------------------------------*/

void SyncSubRun(SyncSub *ssub, const char *caller,
                const Int32 *pPos, const Int32 *pCorr,
                const Int32 *nPos, const Int32 *nCorr)
{
  /* initialize */
  Int16 ppPeaks = 0;
  Int16 nnPeaks = 0;
  Int16 npPeaks = 0;

  /* compute frame numbers for peak positions */
  Int16 p1 = UpdatePeak(pPos, pCorr, -PEAK_DIST_PN - PEAK_DIST_PP);
  Int16 pM = UpdatePeak(pPos, pCorr, -PEAK_DIST_PN - PEAK_DIST_PP/2);
  Int16 p2 = UpdatePeak(pPos, pCorr, -PEAK_DIST_PN);
  Int16 n1 = UpdatePeak(nPos, nCorr, -2*PEAK_DIST_PN - PEAK_DIST_PP);
  Int16 n2 = 8;

  /* target delays */
  Int16 pt = pPos[p2] - 164 + PEAK_DIST_PN;
  Int16 nt = nPos[n2] - 164;

  /* check peak distances */
  if (nPos[n2]-nPos[n1] == PEAK_DIST_NN)   nnPeaks = 2;
  if (pPos[p2]-pPos[p1] == PEAK_DIST_PP)   ppPeaks = 2;
  if (pPos[p2]-pPos[pM] == PEAK_DIST_PP/2) ppPeaks++;

  if (pPos[p1]-nPos[n1] == PEAK_DIST_PN)   npPeaks++;
  if (nPos[n2]-pPos[p2] == PEAK_DIST_PN)   npPeaks++;

  if (nPos[n2]-pPos[pM] == PEAK_DIST_PN + PEAK_DIST_PP/2) npPeaks++;
  if (pPos[pM]-nPos[n1] == PEAK_DIST_PN + PEAK_DIST_PP/2) npPeaks++;

  /* check positive and negative peaks */
  CheckPosPeaks(ssub, caller, pCorr, p1, p2, ppPeaks, npPeaks, pt);
  CheckNegPeaks(ssub, caller, nCorr, n1, n2, nnPeaks, npPeaks, nt);
}

/*============================================================================*/
/* UTILITY FUNCTION: UpdatePeak                                               */
/*----------------------------------------------------------------------------*/
/* Description: update sync peak position                                     */
/*                                                                            */
/* In:     const Int32* pos   -> vector of positions                          */
/*         const Int32* corr  -> vector of correlation values                 */
/*         Int16        dist  -> distance to be checked                       */
/* Return: Int16              <- updated peak position                        */
/*----------------------------------------------------------------------------*/

Int16 UpdatePeak(const Int32 *pos, const Int32 *corr, Int16 dist)
{
  Int16 i = dist + PCM_LENGTH;
  Int16 j = dist + 1;
  Int16 peak;

  if (pos[7+i/PCM_LENGTH] <= i) {
    if (pos[7+j/PCM_LENGTH] < j) {
      peak = 7+i/PCM_LENGTH;
    } else {
      peak = (corr[7+j/PCM_LENGTH] < corr[7+i/PCM_LENGTH]) ?
        7+i/PCM_LENGTH : 7+j/PCM_LENGTH;
    }
  } else {
    peak = (pos[7+j/PCM_LENGTH] < j) ? 0 : 7+j/PCM_LENGTH;
  }
  return peak;
}

/*============================================================================*/
/* UTILITY FUNCTION: CheckPosPeaks                                            */
/*----------------------------------------------------------------------------*/
/* Description: check positive sync peaks                                     */
/*                                                                            */
/* InOut:  SyncSub*     ssub         <-> sync subsystem                       */
/* In:     const char*  caller        -> modem identification                 */
/*         const Int32* pCorr         -> vector of correlation values         */
/*         Int16        p1            -> peak position p1                     */
/*         Int16        p2            -> peak position p2                     */
/*         Int16        ppPeaks       -> number correct pos/pos distances     */
/*         Int16        npPeaks       -> number correct neg/pos distances     */
/*         Int16        targetDelay   -> target delay if sync successful      */
/*----------------------------------------------------------------------------*/

void CheckPosPeaks(SyncSub *ssub, const char *caller, const Int32 *pCorr,
                   Int16 p1, Int16 p2, Int16 ppPeaks, Int16 npPeaks,
                   Int16 targetDelay)
{
  Int16 i, cond1, cond2, jmax = 0;
  Int32 temp;

  /* assign positive dummy value (actual sync delay will be negative) */
  ssub->delay = 1;

  if (p2 != 0) {
    for (i = 1; i < 9; i++) {
      if (pCorr[i] >= MAX(pCorr[p2], pCorr[p1])) jmax++;
    }
  }
  if (p2 != 0 && jmax <= 3) {
    if (ssub->flag == False && ssub->trials < 16 && ppPeaks >= 2
      && pCorr[p1] > SYNC_THRESHOLD
      && pCorr[p2] > SYNC_THRESHOLD)
    {
      ssub->delay = targetDelay;

      if (ssub->amplitude != NULL) {
        temp = pCorr[p1]/2 + pCorr[p2]/2;

        cond1 =
          ssub->amplitude[0] < temp &&
          ssub->amplitude[1]/2 < temp &&
          ssub->npeaks < ppPeaks + npPeaks &&
          MAX(pCorr[p1], pCorr[p2])/3 < MIN(pCorr[p1], pCorr[p2]);

        cond2 = ssub->amplitude[0] < temp && ppPeaks == 3 && npPeaks >= 1;

        if (cond1 || cond2) {
          ssub->flag = True;
          ssub->npeaks = ppPeaks + npPeaks;
          ssub->amplitude[0] = temp;
        }
      } else {
        ssub->flag = True;
      }
      if (ssub->flag == True) {
        LogInfo("[%-4s] sync detected; delay: %+4d (positive peaks)",
          caller, ssub->delay);
      }
    }
  }
}

/*============================================================================*/
/* UTILITY FUNCTION: CheckNegPeaks                                            */
/*----------------------------------------------------------------------------*/
/* Description: check negative sync peaks                                     */
/*                                                                            */
/* InOut:  SyncSub*     ssub         <-> sync subsystem                       */
/* In:     const char*  caller        -> modem identification                 */
/*         const Int32* nCorr         -> vector of correlation values         */
/*         Int16        n1            -> peak position n1                     */
/*         Int16        n2            -> peak position n2                     */
/*         Int16        nnPeaks       -> number correct neg/neg distances     */
/*         Int16        npPeaks       -> number correct neg/pos distances     */
/*         Int16        targetDelay   -> target delay if sync successful      */
/*----------------------------------------------------------------------------*/

void CheckNegPeaks(SyncSub *ssub, const char *caller, const Int32 *nCorr,
                   Int16 n1, Int16 n2, Int16 nnPeaks, Int16 npPeaks,
                   Int16 targetDelay)
{
  Int16 i, cond1, cond2, jmax = 0;
  Int32 temp;

  for (i = 0; i < 9; i++) {
    if (nCorr[i] >= MAX(nCorr[n2], nCorr[n1])) jmax++;
  }
  if (jmax <= 2) {
    if (ssub->flag == False && nnPeaks >= 2
      && nCorr[n1] > SYNC_THRESHOLD
      && nCorr[n2] > SYNC_THRESHOLD)
    {
      if (ssub->amplitude != NULL) {
        temp = nCorr[n1]/2 + nCorr[n2]/2;

        cond1 =
          ssub->amplitude[0] < temp &&
          ssub->amplitude[1]/2 < temp &&
          ssub->npeaks < nnPeaks + npPeaks &&
          MAX(nCorr[n1], nCorr[n2])/3 < MIN(nCorr[n1], nCorr[n2]);

        cond2 = ssub->amplitude[0] < temp && npPeaks > 2;

        if (cond1 || cond2 ||
          (ssub->amplitude[0] < temp && ssub->delay == targetDelay))
        {
          ssub->flag = True;
          ssub->delay = targetDelay;
          ssub->npeaks = nnPeaks + npPeaks;
          ssub->amplitude[0] = temp;
        }
      } else {
        ssub->flag = True;
        ssub->delay = targetDelay;
      }
      if (ssub->flag == True) {
        LogInfo("[%-4s] sync detected; delay: %+4d (negative peaks)",
          caller, ssub->delay);
      }
    }
  }
}

/*============================================================================*/
/* UTILITY FUNCTION: ToneDetect                                               */
/*----------------------------------------------------------------------------*/
/* Description: tone detection at 500 Hz or 800 Hz                            */
/*                                                                            */
/* InOut:  SyncState*   sync  <-> sync struct                                 */
/* In:     const Int16* pcm    -> input frame                                 */
/*----------------------------------------------------------------------------*/

void ToneDetect(SyncState *sync, const Int16 *pcm)
{
  Int16 i;
  Int32 sumSin500 = 0;
  Int32 sumCos500 = 0;
  Int32 sumSin800 = 0;
  Int32 sumCos800 = 0;

  if (sync->memWakeup != NULL) {
    for (i = 0; i < NRF_SYNC; i++) {
      sync->memWakeup[2*i]   = sync->memWakeup[2*(i+1)];
      sync->memWakeup[2*i+1] = sync->memWakeup[2*(i+1)+1];
    }
    for (i = 0; i < PCM_LENGTH; i++) {
      sumSin500 += (wakeupSin500[i%16] >> 5)*pcm[i];
      sumCos500 += (wakeupCos500[i%16] >> 5)*pcm[i];
      sumSin800 += (wakeupSin800[i%10] >> 5)*pcm[i];
      sumCos800 += (wakeupCos800[i%10] >> 5)*pcm[i];
    }
    sync->memWakeup[2*NRF_SYNC] =
      (sumSin500 >> 18)*(sumSin500 >> 18) +
      (sumCos500 >> 18)*(sumCos500 >> 18);

    sync->memWakeup[2*NRF_SYNC+1] =
      (sumSin800 >> 18)*(sumSin800 >> 18) +
      (sumCos800 >> 18)*(sumCos800 >> 18);
  }
}
