/*============================================================================*/
/* eCall ANSI C fixed-point reference source code                             */
/*                                                                            */
/* File:         modem_ivs.c                                                  */
/* Version:      8.4.0 (Rel8) / 9.2.0 (Rel9)                                  */
/* Date:         2010-06-15                                                   */
/* Description:  eCall IVS modem                                              */
/*----------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>

#include "modemx.h"
#include "ecall_defines.h"
#include "ecall_control.h"
#include "ecall_modem.h"
#include "ecall_sync.h"
#include "ecall_fec.h"
#include "ecall_rom.h"


static IvsData ivs;

/* IVS external reference */
IvsData* IvsGetData(void) { return &ivs; }

/*============================================================================*/
/* IVS IMPLEMENTATION: IvsReset, IvsSendStart                                 */
/*----------------------------------------------------------------------------*/
/* In:  const Ord8* msd     -> MSD to be transmitted                          */
/*      int         length  -> MSD length (equal to MSD_MAX_LENGTH)           */
/*----------------------------------------------------------------------------*/

void IvsReset(const Ord8 *msd, int length) {
  IvsRxReset();
  IvsTxReset(msd, length);
}

void IvsSendStart(void) {
  if (ivs.rx.state == IvsIdle) {
    ivs.rx.state = IvsTrigger;
  }
  LogInfo("[IVS ] SendStart received from control\n----");
}

/*----------------------------------------------------------------------------*/

void IvsRxReset(void) {
  CtrlRxData *ctrl = &ivs.rx.ctrl;
  SyncState  *sync = &ivs.rx.sync;

  /* reset state */
  ivs.rx.state = IvsIdle;

  /* reset control receiver */
  CtrlRxReset(ctrl, "IVS", sync, ivs.rx.memCtrl, LOCK_START_DL);

  /* reset counters for SEND uplink control */
  ivs.rx.dlData   = DlMsgNoop;
  ivs.rx.dlMsgCnt = DlCntStart;
  ivs.rx.dlIndex  = -1;

  /* reset sync */
  SyncReset(sync, ivs.rx.memSync, NULL);

  LogInfo("[IVS ] reset receiver");
}

/*----------------------------------------------------------------------------*/

void IvsTxReset(const Ord8 *msd, int length) {
  Int16 k;

  /* reset state */
  ivs.tx.state = IvsIdle;

  /* set control link parameters (SEND) */
  CtrlTxReset(&ivs.tx.ctrl, "IVS");

  /* reset state counters */
  for (k = 0; k <= SET_LLMSG; k++) {
    ivs.tx.stateCnt[k] = 0;
    ivs.tx.stateIgn[k] = 0;
  }
  ivs.tx.stateCnt[DlMsgSend] = -1;
  ivs.tx.startPending = False;

  /* reset uplink counters */
  ivs.tx.rv      = 0;
  ivs.tx.ulN     = 0;
  ivs.tx.ulIndex = 0;
  ivs.tx.ulDelay = 73;

  if (msd != NULL) {
    /* initialize modulator to be undefined */
    SetModState(&ivs.tx.mod, ModUndef);

    /* clear global NACK count for new message */
    ivs.tx.stateCntNack = 0;

    /* convert MSD and encode (all redundancy versions) */
    Byte2Bit(msd, ivs.tx.memCode, MSD_MAX_LENGTH);
    FecEncode(ivs.tx.memCode);

    LogInfo("[IVS ] reset transmitter (MSD provided)");
  } else {
    LogInfo("[IVS ] reset transmitter");
  }
}

/*============================================================================*/
/* IVS IMPLEMENTATION: IvsProcess                                             */
/*----------------------------------------------------------------------------*/
/* InOut:  Int16* pcm  <-> input and output frame of 16bit PCM samples        */
/*----------------------------------------------------------------------------*/

void IvsProcess(Int16 *pcm) {
  Int16 i;
  CtrlPort *prx = &ivs.rx.ctrl.port;
  CtrlPort *ptx = &ivs.tx.ctrl.port;

  /* invert input signal if required */
  if (prx->invert == True) {
    for (i = 0; i < PCM_LENGTH; i++) {
      pcm[i] = -pcm[i];
    }
  }
  IvsRxProcess(pcm);

  if (prx->state == CtrlTxIdle || prx->state == CtrlTxSend) {
    /* control transmit mode */
    ptx->state        = prx->state;
    ptx->u.tx.dlData  = prx->u.tx.dlData;
    ptx->u.tx.dlIndex = prx->u.tx.dlIndex;
  }
  else {
    /* control receive mode */
    ptx->state         = prx->state;
    ptx->u.rx.dlData   = prx->u.rx.dlData;
    ptx->u.rx.dlMetric = prx->u.rx.dlMetric;
  }
  IvsTxProcess(pcm);

  /* invert output signal if required */
  if (ptx->invert == True) {
    for (i = 0; i < PCM_LENGTH; i++) {
      pcm[i] = -pcm[i];
    }
  }
}

/*----------------------------------------------------------------------------*/

void IvsRxProcess(const Int16 *pcm) {
  Int16 i;
  CtrlPort *port = &ivs.rx.ctrl.port;

  /* IVS receiver, copy PCM data to input buffer */
  for (i = 0; i < NRF_MEMCTRL-1; i++) {
    memcpy(&ivs.rx.memCtrl[i*PCM_LENGTH],
      &ivs.rx.memCtrl[(i+1)*PCM_LENGTH], sizeof(Int16)*PCM_LENGTH);
  }
  memcpy(&ivs.rx.memCtrl[(NRF_MEMCTRL-1)*PCM_LENGTH], pcm,
    sizeof(Int16)*PCM_LENGTH);

  /* IVS receiver, run control message receiver */
  CtrlRxProcess(&ivs.rx.ctrl, pcm);

  /* check for uplink SEND trigger */
  switch (ivs.rx.state) {
    case IvsTrigger:
      if (port->state == CtrlRxIdle) {
        port->state = CtrlTxIdle;

        /* set control data */
        ivs.rx.state    = IvsStart;
        ivs.rx.dlData   = DlMsgSend;
        ivs.rx.dlMsgCnt = DlCntNext;
        ivs.rx.dlIndex  = 0;
      } else {
        /* SEND trigger not allowed during MSD transmission */
        ivs.rx.state = IvsIdle;
      }
      break;
    case IvsStart:
      if (port->state == CtrlRxSync || ivs.rx.dlMsgCnt >= IVS_NUMSEND) {
        /* stop SEND trigger */
        ivs.rx.state    = IvsIdle;
        ivs.rx.dlData   = DlMsgNoop;
        ivs.rx.dlMsgCnt = DlCntWait;
        ivs.rx.dlIndex  = -1;
      } else {
        /* sending control data */
        port->state        = CtrlTxSend;
        port->u.tx.dlData  = ivs.rx.dlData;
        port->u.tx.dlIndex = ivs.rx.dlIndex;

        if (ivs.rx.dlIndex++ == 0) {
          LogInfo("[IVS ] sending  SEND");
        }
        else if (ivs.rx.dlIndex == NRF_DLCHUNK) {
          ivs.rx.dlIndex = 0;
          ivs.rx.dlMsgCnt++;
        }
      }
    default:
      /* do nothing but receiving downlink data */
      break;
  }
}

/*----------------------------------------------------------------------------*/

void IvsTxProcess(Int16 *pcm) {
  Int16 lpcm = sizeof(Int16)*PCM_LENGTH;
  CtrlTxData *ctrl = &ivs.tx.ctrl;

  /* IVS transmitter, initialize PCM data to zero */
  memset(pcm, 0, lpcm);

  if (ctrl->port.state == CtrlTxIdle) {
    /* SEND trigger initiated */
  }
  else if (ctrl->port.state == CtrlTxSend) {
    /* transmit control data on uplink */
    CtrlTxProcess(ctrl, pcm);
  }
  else {
    /* consecutive sync check failures may trigger a reset */
    if (ctrl->port.u.rx.dlData == DlMsgReset) {
      if (ivs.tx.state == IvsAck ||
        (ivs.tx.rv == 0 && ivs.tx.ulIndex == 0) || ivs.tx.rv > ARQ_MAX-1)
      {
        IvsReset(NULL, 0);
        LogInfo("[IVS ] full restart (too many sync check failures)");
      }
    }
    else if (ctrl->port.u.rx.dlData != DlMsgNoop) {
      /* run state machine based on new downlink symbol */
      IvsTxState(ctrl->port.u.rx.dlData, ctrl->port.u.rx.dlMetric,
        (ctrl->port.state == CtrlRxLock) ? True : False);
    }

    /* run transmitter based on IVS state */
    if (ivs.tx.state == IvsIdle || ivs.tx.state == IvsAck) {
      /* do nothing */
    }
    else if (ivs.tx.state == IvsSendMsd && ivs.tx.mod.type != ModUndef) {
      if (ivs.tx.ulIndex < ivs.tx.ulN) {
        /* run transmitter */
        IvsTransmitter(
          &ivs.tx.mod, ivs.tx.memCode, pcm, ivs.tx.rv, ivs.tx.ulIndex);
        ivs.tx.ulIndex++;
      }
      else {
        /* increase redundancy version counter */
        ivs.tx.rv++;
        ivs.tx.ulIndex = 0;

        /* no sync frame is transmitted next */
        ivs.tx.ulN = ivs.tx.mod.nfmuteall + ivs.tx.mod.nfdata;

        /* new redundancy version, run transmitter */
        IvsTransmitter(
          &ivs.tx.mod, ivs.tx.memCode, pcm, ivs.tx.rv, ivs.tx.ulIndex);
        ivs.tx.ulIndex++;
      }
      /* randomize transmit start */
      memcpy(ivs.tx.memDelay, &ivs.tx.memDelay[PCM_LENGTH], lpcm);
      memcpy(&ivs.tx.memDelay[PCM_LENGTH], pcm, lpcm);
      memcpy(pcm, &ivs.tx.memDelay[PCM_LENGTH - ivs.tx.ulDelay], lpcm);
    }
  }
}

/*============================================================================*/
/* UTILITY FUNCTION: IvsTxState                                               */
/*----------------------------------------------------------------------------*/
/* In:  Int16 msg       -> new control message symbol                         */
/*      Int16 metric    -> receiver metric (-1: ignore symbol)                */
/*      Bool  syncLock  -> indicates sync lock of control receiver            */
/*----------------------------------------------------------------------------*/

void IvsTxState(Int16 msg, Int16 metric, Bool syncLock) {
  Int16 k;

  if (syncLock == True) {
    ivs.tx.startPending = True;
  }
  if (msg >= SET_LLMSG && msg < SET_LLMSG + SET_HLMSG) {
    /* higher-layer message symbol */
    msg -= SET_LLMSG;

    if (ivs.tx.stateCnt[DlMsgHlack] + ivs.tx.stateIgn[DlMsgHlack] > 0 &&
      ivs.rx.ctrl.dlHlackData == (Ord8)msg)
    {
      if (metric < 0)
        ivs.tx.stateIgn[DlMsgHlack]++;
      else
        ivs.tx.stateCnt[DlMsgHlack]++;
    }
    else if (ivs.tx.stateCnt[DlMsgHlack] == 0 && metric < 0) {
      ivs.rx.ctrl.dlHlackData = (Ord8)msg;
      ivs.tx.stateIgn[DlMsgHlack] = 1;
    }
    else if (metric >= 0) {
      ivs.rx.ctrl.dlHlackData = (Ord8)msg;
      ivs.tx.stateCnt[DlMsgHlack] = 1;
      ivs.tx.stateIgn[DlMsgHlack] = 0;
    }

    /* two conditions for successful HLACK reception */
    if (
      ivs.tx.stateCnt[DlMsgHlack] >= 2 ||
      ivs.tx.stateCnt[DlMsgHlack] + ivs.tx.stateIgn[DlMsgHlack] >= 3)
    {
      LogInfo("[IVS ] received HLACK (metric: %3d), finished.", metric);

      /* invoke callback function and stop */
      IvsReceiveHlack(ivs.rx.ctrl.dlHlackData);
      ivs.tx.state = IvsIdle;

      /* entering IDLE state -> reset all counters */
      for (k = 0; k <= SET_LLMSG; k++) {
        ivs.tx.stateCnt[k] = 0;
        ivs.tx.stateIgn[k] = 0;
      }
      ivs.tx.stateCnt[DlMsgSend] = -1;
    } else {
      LogInfo("[IVS ] received HLACK (metric: %3d), waiting...", metric);
    }
  } else {
    /* lower-layer message symbol */
    switch (msg) {

      case DlMsgStart:
        if (metric < 0 && ivs.tx.stateIgn[DlMsgStart] < IVS_GOSTART) {
          /* ignore, message reliability too low */
          ivs.tx.stateIgn[DlMsgStart]++;
          LogInfo("[IVS ] received START !> ignoring message (%s)",
            "reliability too low");
          break;
        }
        if (ivs.tx.startPending == True) {
          ivs.tx.startPending = False;
          ivs.tx.stateCnt[DlMsgStart] = LOCK_START_DL;
        } else {
          ivs.tx.stateCnt[DlMsgStart]++;
        }

        /* check start... */
        if (ivs.tx.stateCnt[DlMsgStart] > LOCK_START_DL + FAIL_RESTART &&
          (ivs.tx.stateCnt[DlMsgStart]-1)%FAIL_RESTART == 0)
        {
          /* assume PSAP sync failure, starting over... */
          ivs.tx.ulDelay = (ivs.tx.ulDelay + IVS_TXINC)%PCM_LENGTH;
          ivs.tx.ulN     = NRF_SYNC + ivs.tx.mod.nfmuteall + ivs.tx.mod.nfdata;
          ivs.tx.ulIndex = 0;
          ivs.tx.rv      = 0;

          /* clear counter for unreliable START messages */
          ivs.tx.stateIgn[DlMsgStart] = 0;
          LogInfo("[IVS ] received START (metric: %3d) too often, %s", metric,
            "starting over...");
        }
        else if (ivs.tx.stateCnt[DlMsgStart] == LOCK_START_DL) {
          /* downlink sync locked, start sending */
          ivs.tx.state = IvsSendMsd;
          LogInfo("[IVS ] received START (metric: %3d), starting...", metric);

          /* choose type of uplink modulator */
          if (ivs.tx.stateCntNack < IVS_TXFAST) {
            SetModState(&ivs.tx.mod, Mod3bit4smp);
            LogInfo("[IVS ] fast modulator chosen (NACK count: %d)\n----",
              ivs.tx.stateCntNack);
          } else {
            SetModState(&ivs.tx.mod, Mod3bit8smp);
            LogInfo("[IVS ] robust modulator chosen (NACK count: %d)\n----",
              ivs.tx.stateCntNack);
          }
          /* define uplink parameters */
          ivs.tx.ulDelay = (ivs.tx.ulDelay + IVS_TXINC)%PCM_LENGTH;
          ivs.tx.ulN     = NRF_SYNC + ivs.tx.mod.nfmuteall + ivs.tx.mod.nfdata;
          ivs.tx.ulIndex = 0;
          ivs.tx.rv      = 0;

          /* entering new state -> reset other counters */
          ivs.tx.stateIgn[DlMsgStart] = 0;
          ivs.tx.stateCnt[DlMsgNack]  = 0;
          ivs.tx.stateCnt[DlMsgAck]   = 0;
          ivs.tx.stateCnt[DlMsgSend]  = 0;
        }
        else {
          /* preserve state */
          LogInfo("[IVS ] received START (metric: %3d)", metric);
        }
        break;

      case DlMsgNack:
        if (ivs.tx.startPending == True) {
          /* downlink sync locked, still waiting to start */
          LogInfo("[IVS ] received NACK  !> ignoring message (%s)",
            "START trigger pending");
          break;
        }
        else if (metric < 0) {
          /* ignore, message reliability too low */
          LogInfo("[IVS ] received NACK  !> ignoring message (%s)",
            "reliability too low");
          break;
        }
        /* message seems reliable, increase counters */
        ivs.tx.stateCnt[DlMsgNack]++;
        ivs.tx.stateCntNack++;
        if (
          ivs.tx.stateCnt[DlMsgNack] == 1 &&
          ivs.tx.stateCnt[DlMsgSend] < 0)
        {
          /* ignore, NACK should only follow START or NACK */
          ivs.tx.stateCnt[DlMsgNack] = 0;
          LogInfo("[IVS ] received NACK  !> ignoring message (%s)",
            "assume IDLE");
        }
        else if (
          ivs.tx.stateCnt[DlMsgNack] == 1 &&
          ivs.tx.stateCnt[DlMsgAck] > 0)
        {
          /* NACK after single ACK -> assume NACK */
          ivs.tx.stateCnt[DlMsgAck] = 0;
          LogInfo("[IVS ] received NACK  !> probably ACK was wrong");
        }
        else if (ivs.tx.stateCnt[DlMsgNack] == 1) {
          LogInfo("[IVS ] received NACK  (metric: %3d), first time", metric);

          /* entering new state -> reset other counters */
          ivs.tx.stateIgn[DlMsgStart] = 0;
          ivs.tx.stateCnt[DlMsgStart] = 0;
          ivs.tx.stateCnt[DlMsgAck]   = 0;
          ivs.tx.stateCnt[DlMsgSend]  = 0;
        }
        else {
          /* preserve state */
          LogInfo("[IVS ] received NACK  (metric: %3d)", metric);
        }
        break;

      case DlMsgAck:
        if (ivs.tx.startPending == True) {
          /* downlink sync locked, still waiting to start */
          LogInfo("[IVS ] received ACK   !> ignoring message (%s)",
            "START trigger pending");
          break;
        }
        else if (metric < 0) {
          /* ignore, message reliability too low */
          LogInfo("[IVS ] received ACK   !> ignoring message (%s)",
            "reliability too low");
          break;
        }
        /* message seems reliable, two ACK messages required */
        ivs.tx.stateCnt[DlMsgAck]++;
        if (
          ivs.tx.stateCnt[DlMsgAck] == 1 &&
          ivs.tx.stateCnt[DlMsgSend] < 0)
        {
          /* ignore, IDLE mode precedes ACK */
          ivs.tx.stateCnt[DlMsgAck] = 0;
          LogInfo("[IVS ] received ACK   !> ignoring message (%s)",
            "assume IDLE");
        }
        else if (ivs.tx.stateCnt[DlMsgAck] == 2) {
          LogInfo("[IVS ] received ACK   (metric: %3d), %s", metric,
            "ACK recorded, wait for HLACK");

          /* invoke callback function, enter ACK state, wait for HLACK */
          IvsReceiveAck();
          ivs.tx.state = IvsAck;
        }
        else {
          /* waiting... */
          LogInfo("[IVS ] received ACK   (metric: %3d), waiting...", metric);
        }
        break;

      default:
        if (ivs.tx.startPending == True) {
          /* downlink sync locked, still waiting to start */
          LogInfo("[IVS ] received IDLE  !> ignoring message (%s)",
            "START trigger pending");
        }
        else if (
          ivs.tx.stateCnt[DlMsgStart] > 0 ||
          ivs.tx.stateCnt[DlMsgNack] > 0 ||
          ivs.tx.stateCnt[DlMsgAck] > 0)
        {
          LogInfo("[IVS ] received IDLE  !> ignoring message");
        }
        else {
          LogInfo("[IVS ] remaining in IDLE mode");
        }
        break;
    }
  }
}
