/*============================================================================*/
/* eCall ANSI C fixed-point reference source code                             */
/*                                                                            */
/* File:         modem_psap.c                                                 */
/* Version:      8.4.0 (Rel8) / 9.2.0 (Rel9)                                  */
/* Date:         2010-06-15                                                   */
/* Description:  eCall PSAP modem                                             */
/*----------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>

#include "modemx.h"
#include "ecall_defines.h"
#include "ecall_control.h"
#include "ecall_modem.h"
#include "ecall_sync.h"
#include "ecall_fec.h"
#include "ecall_rom.h"


static PsapData psap;

/* PSAP external reference */
PsapData* PsapGetData(void) { return &psap; }

/*============================================================================*/
/* PSAP IMPLEMENTATION: PsapReset, PsapSendStart, PsapSendHlack               */
/*----------------------------------------------------------------------------*/

void PsapReset(void) {
  PsapRxReset();
  PsapTxReset();
}

void PsapSendStart(void) {
  PsapRxReset();
  psap.rx.state = PsapTrigger;

  LogInfo("[PSAP] SendStart received from control\n----");
}

void PsapSendHlack(const Ord8 data) {
  /* HLACK mode requires MSD to be successfully received */
  if (psap.rx.state == PsapAck) {
    psap.rx.state       = PsapHlack;
    psap.rx.dlMsgCnt    = DlCntWait;
    psap.rx.dlHlackData = data & 0x0F;

    LogInfo("[PSAP] HLACK data received from control\n----");
  } else {
    LogInfo("[PSAP] HLACK data received from control, ignored.\n----");
  }
}

/*----------------------------------------------------------------------------*/

void PsapRxReset(void) {
  CtrlRxData *ctrl = &psap.rx.ctrl;
  SyncState  *sync = &psap.rx.sync;

  /* reset state */
  psap.rx.state = PsapIdle;

  /* assign buffer pointers */
  psap.rx.msdBin  = (Ord1*)  &psap.rx.buffer[0];
  psap.rx.msd     = (Ord8*)  &psap.rx.buffer[NRB_INFO_CRC];
  psap.rx.memCtrl = (Int16*) &psap.rx.buffer[sizeof(IntLLR)*NRB_CODE_ARQ];
  psap.rx.memCode = (IntLLR*)&psap.rx.buffer[0];

  /* reset control receiver */
  CtrlRxReset(ctrl, "PSAP", sync, psap.rx.memCtrl, LOCK_START_UL);

  /* reset uplink counters */
  psap.rx.rv         = 0;
  psap.rx.ulIndex    = 0;
  psap.rx.mgIndex    = 0;
  psap.rx.ulTrials   = 0;
  psap.rx.ulSyncTail = 0;

  /* reset HLACK data and downlink counters */
  psap.rx.dlHlackData = 0;
  psap.rx.dlData      = DlMsgNoop;
  psap.rx.dlMsgCnt    = DlCntStart;
  psap.rx.dlIndex     = -1;

  /* reset sync */
  SyncReset(sync,
    (Int32*)&psap.rx.buffer[0
    + sizeof(IntLLR)* NRB_CODE_ARQ
    + sizeof(Int16) * NRF_MEMCTRL*PCM_LENGTH],
    (Int32*)&psap.rx.buffer[0
    + sizeof(IntLLR)* NRB_CODE_ARQ
    + sizeof(Int16) * NRF_MEMCTRL*PCM_LENGTH
    + sizeof(Int32) * NRS_MEMSYNC]);

  /* clear buffer */
  memset(psap.rx.buffer, 0, sizeof(char)*(0
    + sizeof(IntLLR)* NRB_CODE_ARQ
    + sizeof(Int16) * NRF_MEMCTRL*PCM_LENGTH
    + sizeof(Int32) * NRS_MEMSYNC
    + sizeof(Int32) * 2*(NRF_SYNC+1)));

  LogInfo("[PSAP] reset receiver");
}

/*----------------------------------------------------------------------------*/

void PsapTxReset(void) {
  CtrlTxReset(&psap.tx.ctrl, "PSAP");

  LogInfo("[PSAP] reset transmitter");
}

/*============================================================================*/
/* PSAP IMPLEMENTATION: PsapProcess                                           */
/*----------------------------------------------------------------------------*/
/* InOut:  Int16* pcm  <-> input and output frame of 16bit PCM samples        */
/*----------------------------------------------------------------------------*/

void PsapProcess(Int16 *pcm) {
  Int16 i;
  CtrlPort *prx = &psap.rx.ctrl.port;
  CtrlPort *ptx = &psap.tx.ctrl.port;

  /* invert input signal if required */
  if (prx->invert == True) {
    for (i = 0; i < PCM_LENGTH; i++) {
      pcm[i] = -pcm[i];
    }
  }
  PsapRxProcess(pcm);

  if (prx->state == CtrlTxIdle || prx->state == CtrlTxSend) {
    /* port match, interface copy */
    ptx->state        = prx->state;
    ptx->u.tx.dlData  = prx->u.tx.dlData;
    ptx->u.tx.dlIndex = prx->u.tx.dlIndex;
  }
  else {
    /* port mismatch, no downlink payload */
    ptx->state        = CtrlTxIdle;
    ptx->u.tx.dlData  = DlMsgNoop;
    ptx->u.tx.dlIndex = -1;
  }
  PsapTxProcess(pcm);

  /* invert output signal if required */
  if (ptx->invert == True) {
    for (i = 0; i < PCM_LENGTH; i++) {
      pcm[i] = -pcm[i];
    }
  }
}

/*----------------------------------------------------------------------------*/

void PsapRxProcess(const Int16 *pcm) {
  Int16 i;

  /* PSAP receiver, copy PCM data to input buffer */
  for (i = 0; i < NRF_MEMCTRL-1; i++) {
    memcpy(&psap.rx.memCtrl[i*PCM_LENGTH],
      &psap.rx.memCtrl[(i+1)*PCM_LENGTH], sizeof(Int16)*PCM_LENGTH);
  }
  memcpy(&psap.rx.memCtrl[(NRF_MEMCTRL-1)*PCM_LENGTH], pcm,
    sizeof(Int16)*PCM_LENGTH);

  /* process uplink and downlink state machine */
  PsapRxUplink(pcm);
  PsapRxDownlink();
}

/*----------------------------------------------------------------------------*/

void PsapTxProcess(Int16 *pcm) {
  CtrlTxProcess(&psap.tx.ctrl, pcm);
}

/*============================================================================*/
/* UTILITY FUNCTION: PsapRxUplink                                             */
/*----------------------------------------------------------------------------*/
/* In:  const Int16* pcm  -> input frame of 16bit PCM samples                 */
/*----------------------------------------------------------------------------*/

void PsapRxUplink(const Int16 *pcm) {
  Int16 temp;
  Int32 xCorrA, xCorrB;
  Bool  crc, decode;

  CtrlRxData *ctrl = &psap.rx.ctrl;
  SyncState  *sync = &psap.rx.sync;
  ModState   *mod  = &psap.rx.mod;

  /* process each state individually */
  switch (psap.rx.state) {
    case PsapIdle:
      /* run control receiver and watch for SEND signal */
      CtrlRxProcess(ctrl, pcm);

      if (ctrl->port.u.rx.dlData == DlMsgSend) {
        LogInfo("[PSAP] received SEND  (metric: %3d), %s",
          ctrl->port.u.rx.dlMetric, "initiating START trigger");

        /* reset receiver and prepare for MSD reception */
        PsapRxReset();
        psap.rx.state = PsapTrigger;
      }
      break;

    case PsapTrigger:
      /* start sending control messages, wait for IVS to transmit */
      ctrl->port.state = CtrlTxIdle;
      break;

    case PsapStart:
      /* run sync with new input data */
      ctrl->port.state = CtrlTxIdle;
      Sync(sync, pcm, "PSAP", ctrl->port.invert);

      if (sync->flag == True && sync->delay <= 0) {
        /* sync successful */
        ctrl->port.state = CtrlTxSend;

        /* apply sync inversion flag */
        if ((ctrl->port.invert = sync->invert) == True) {
          LogInfo("[PSAP] codec inversion detected!\n----");
        }
        /* start sending NACK messages */
        psap.rx.state    = PsapNack;
        psap.rx.dlMsgCnt = DlCntWait;

        /* choose modulator */
        xCorrA =
          sync->memWakeup[0] +
          sync->memWakeup[2] +
          sync->memWakeup[4] +
          sync->memWakeup[6];
        xCorrB =
          sync->memWakeup[1] +
          sync->memWakeup[3] +
          sync->memWakeup[5] +
          sync->memWakeup[7];

        if (xCorrA/MAX(xCorrB, 1) > PSAP_THRESHOLD) {
          SetModState(mod, Mod3bit4smp);
          LogInfo("[PSAP] fast modulator chosen (metric: %d)",
            xCorrA/MAX(xCorrB, 1));
        }
        else if (xCorrB/MAX(xCorrA, 1) > PSAP_THRESHOLD) {
          SetModState(mod, Mod3bit8smp);
          LogInfo("[PSAP] robust modulator chosen (metric: %d)",
            xCorrB/MAX(xCorrA, 1));
        }
        else if (psap.rx.ulTrials == 0) {
          SetModState(mod, Mod3bit4smp);
          LogInfo("[PSAP] fast modulator chosen (metric: %d)",
            xCorrA/MAX(xCorrB, 1));
        }
        else {
          SetModState(mod, Mod3bit8smp);
          LogInfo("[PSAP] robust modulator chosen (metric: %d)",
            xCorrB/MAX(xCorrA, 1));
        }
        psap.rx.ulN = mod->nfdata;

        /* check for muting gaps passed already */
        if (sync->delay < (-PCM_LENGTH)) {
          temp = MIN(-(sync->delay)/PCM_LENGTH, mod->nfmute1);
          sync->delay += temp*PCM_LENGTH;
          psap.rx.mgIndex += temp;
        }
      }
      break;

    case PsapNack:
      /* check for better sync for some more frames */
      if (psap.rx.ulSyncTail < NRF_OBSERVE) {
        psap.rx.ulSyncTail++;

        /* run synchronization */
        Sync(sync, pcm, "PSAP", ctrl->port.invert);

        if (sync->flag == True && sync->delay <= 0) {
          /* apply sync inversion flag */
          if ((ctrl->port.invert = sync->invert) == True) {
            LogInfo("[PSAP] codec inversion detected (sync observer)!\n----");
          }
          /* reset receiver, previous sync was likely too early */
          psap.rx.rv         = 0;
          psap.rx.ulIndex    = 0;
          psap.rx.mgIndex    = -1;
          psap.rx.ulSyncTail = 0;

          /* reinitialize buffer */
          memset(psap.rx.memCode, 0, sizeof(IntLLR)*NRB_CODE_ARQ);

          /* check for muting gaps passed already */
          if (sync->delay < (-PCM_LENGTH)) {
            temp = MIN(-(sync->delay)/PCM_LENGTH, mod->nfmute1);
            sync->delay += temp*PCM_LENGTH;
            psap.rx.mgIndex += temp;
          }
        }
      }
      /* run demodulation if not muting */
      decode = False;

      if (psap.rx.mgIndex >= 0) {
        if (mod->mgTable[psap.rx.mgIndex] >= 0) {
          PsapReceiver(mod,
            &psap.rx.memCtrl[(NRF_MEMCTRL-1)*PCM_LENGTH + sync->delay],
            &psap.rx.memCode[(mod->mfpf*mod->bpsym)*psap.rx.ulIndex]);
          psap.rx.ulIndex++;
          decode = True;
        }
      }
      /* increase muting gap counter */
      psap.rx.mgIndex++;

      /* run decoder if desired */
      if (decode == True && (
        ((psap.rx.rv%ARQ_MAX) > 0 && psap.rx.ulIndex == mod->decpos1) ||
        ((psap.rx.rv%ARQ_MAX) > 0 && psap.rx.ulIndex == mod->decpos2) ||
        psap.rx.ulIndex == psap.rx.ulN))
      {
        /* do not run twice with rv = 0, buffer will be reinitialized */
        crc = FecDecode(psap.rx.memCode, psap.rx.rv, psap.rx.msdBin);

        if (crc == False) {
          /* CRC failed, initialization for new redundancy version */
          LogInfo("[PSAP] CRC failed; rv = %d\n----", psap.rx.rv);

          /* reinitialize buffer */
          memset(psap.rx.memCode, 0, sizeof(IntLLR)*NRB_CODE_ARQ);

          /* full block was received, reinitialization necessary */
          if (psap.rx.ulIndex == psap.rx.ulN) {
            psap.rx.ulIndex = 0;
            psap.rx.rv++;

            if ((psap.rx.rv%ARQ_MAX) == 0) {
              /* too many retransmissions, sync must have failed */
              SyncReset(sync, NULL, NULL);

              psap.rx.state      = PsapTrigger;
              psap.rx.dlMsgCnt   = DlCntWait;
              psap.rx.rv         = 0;
              psap.rx.mgIndex    = 0;
              psap.rx.ulSyncTail = 0;
              psap.rx.ulTrials++;
            } else
              psap.rx.mgIndex = -(mod->nfmute4);
          }
        } else {
          LogInfo("[PSAP] MSD successfully received; %s: %d\n====",
            "redundancy versions", psap.rx.rv);

          /* set ACK and convert binary MSD to ASCII */
          psap.rx.state    = PsapAck;
          psap.rx.dlMsgCnt = DlCntWait;
          Bit2Byte(psap.rx.msdBin, psap.rx.msd, MSD_MAX_LENGTH);

          /* send MSD received to control unit */
          PsapReceiveMsd(psap.rx.msd, MSD_MAX_LENGTH);
        }
      }
      break;

    case PsapAck:
      /* MSD was received, keep sending ACK messages */
      break;
    case PsapHlack:
      /* MSD was received, higher-layer message in progress */
      break;
    default:
      /* error */
      Abort("PSAP state not recognized, exit.");
      break;
  }
}

/*============================================================================*/
/* UTILITY FUNCTION: PsapRxDownlink                                           */
/*----------------------------------------------------------------------------*/

void PsapRxDownlink(void) {
  CtrlPort *port = &psap.rx.ctrl.port;

  if (
    port->state == CtrlRxIdle ||
    port->state == CtrlRxSync ||
    port->state == CtrlRxLock)
  {
    /* PSAP in control receiver mode, port data was set above */
  }
  else if (psap.rx.state == PsapIdle && psap.rx.dlMsgCnt >= DlCntNext) {
    /* final message has been sent, stop output */
    port->u.tx.dlData  = DlMsgNoop;
    port->u.tx.dlIndex = -1;
  }
  else {
    /* interface copy */
    port->u.tx.dlData  = psap.rx.dlData;
    port->u.tx.dlIndex = psap.rx.dlIndex;

    /* increase downlink index */
    psap.rx.dlIndex++;
    psap.rx.dlIndex %= NRF_DLCHUNK;

    if (psap.rx.dlIndex == 0) {
      /* negative counter indicates new message */
      if (psap.rx.dlMsgCnt < DlCntNext) {
        psap.rx.dlMsgCnt = DlCntNext;
      } else {
        psap.rx.dlMsgCnt++;
      }
      /* check message counter and change state */
      switch (psap.rx.state) {
        case PsapTrigger:
          if (psap.rx.dlMsgCnt >= LOCK_START_DL) {
            psap.rx.state    = PsapStart;
            psap.rx.dlMsgCnt = DlCntNext;
          } else {
            psap.rx.dlData = DlMsgStart;
            LogInfo("[PSAP] sending  START");
          }
          break;
        case PsapStart:
          if (psap.rx.dlMsgCnt >= (PSAP_NUMSTART - LOCK_START_DL)) {
            psap.rx.state    = PsapIdle;
            psap.rx.dlMsgCnt = DlCntNext;
          }
          break;
        case PsapAck:
          if (psap.rx.dlMsgCnt >= PSAP_NUMACK) {
            psap.rx.state    = PsapIdle;
            psap.rx.dlMsgCnt = DlCntNext;
          }
          break;
        case PsapHlack:
          if (psap.rx.dlMsgCnt >= PSAP_NUMHLACK) {
            psap.rx.state    = PsapIdle;
            psap.rx.dlMsgCnt = DlCntNext;
          } else {
            if (psap.rx.dlMsgCnt == DlCntNext) {
              LogInfo("[PSAP] sending  HLACK; data: 0x0%X",
                psap.rx.dlHlackData);
            } else {
              LogInfo("[PSAP] sending  HLACK");
            }
            /* add message offset */
            psap.rx.dlData = SET_LLMSG + psap.rx.dlHlackData;
          }
          break;
        default:
          /* other states do not require message counting */
          break;
      }
    }
    else if (psap.rx.dlIndex == NRF_SYNC + NRF_DLMUTE1LL) {
      /* start of downlink data portion, set next message */
      switch (psap.rx.state) {
        case PsapTrigger:
          if (psap.rx.dlMsgCnt == DlCntWait) {
            psap.rx.dlData   = DlMsgStart;
            psap.rx.dlMsgCnt = DlCntNext;
            LogInfo("[PSAP] sending  START (restart)");
          }
          break;
        case PsapStart:
          psap.rx.dlData = DlMsgStart;
          if (psap.rx.dlMsgCnt == DlCntWait) {
            psap.rx.dlMsgCnt = DlCntNext;
          }
          LogInfo("[PSAP] sending  START");
          break;
        case PsapNack:
          psap.rx.dlData = DlMsgNack;
          if (psap.rx.dlMsgCnt == DlCntWait) {
            psap.rx.dlMsgCnt = DlCntNext;
          }
          LogInfo("[PSAP] sending  NACK");
          break;
        case PsapAck:
          psap.rx.dlData = DlMsgAck;
          if (psap.rx.dlMsgCnt == DlCntWait) {
            psap.rx.dlMsgCnt = DlCntNext;
          }
          LogInfo("[PSAP] sending  ACK");
          break;
        case PsapHlack:
          if (psap.rx.dlMsgCnt == DlCntWait) {
            psap.rx.dlData = DlMsgAck;
            LogInfo("[PSAP] sending  ACK   (HLACK pending)");
          }
          break;
        default:
          LogInfo("[PSAP] remaining in IDLE mode");
          break;
      }
    }
  }
}
