/**
 * TestLab main driver: simulation implementation.
 *
 * @copyright 2008, Research In Motion Ltd.
 *
 * @version $Id: //depot/main/dsp/ecall/c/testlab/testlab_sim.c#8 $
 */

#include <stdlib.h>

#include "rimtypes.h"
#include "control.h"
#include "ipc.h"
#include "error.h"
#include "option.h"
#include "testlab.h"

static char *options[] = {
  "--logpcm:%s",
  "--logpcm",
  "--log:%s",
  "--loghlack",
  "--ivspush",
  "--ackstop",
  "--startframe:%s",
  "--startframehlack:%s",
  "--totalframes:%s",
  "--totalframesack:%s",
  "--campaign:%s",
  "--invcodec:%s",
  "--invcodec",
  "-h",
  "--help",
  NULL
};

#define ARG_MAX_LENGTH  (200)
#define PORT_MAX_LENGTH (200)
#define HOST_MAX_LENGTH (200)
#define FILE_MAX_LENGTH (200)

static char p1[ARG_MAX_LENGTH];

static char *pcm = NULL;
static char invCodec[4 + 1] = "none";
static char pcmFile[FILE_MAX_LENGTH + 1] = "";
static char logFile[FILE_MAX_LENGTH + 1] = "";
static char cpnFile[FILE_MAX_LENGTH + 1] = "";

static int logHlack = FALSE;
static int ivsPush  = FALSE;
static int ackStop  = FALSE;

static int startFrame      = 0;
static int startFrameHlack = 0;
static int totalFrames     = 0;
static int totalFramesAck  = 0;

static void Usage(char *msg, ...) {
  va_list args;
  va_start(args, msg);

  if (msg != NULL) {
    fprintf(stderr, "Error: ");
    vfprintf(stderr, msg, args);
    fprintf(stderr, "\n");
  }

  fprintf(stderr, "Usage:\n");
  fprintf(stderr, "  testsim_demo <options>\n");
  fprintf(stderr, "    with <options> in:\n");
  fprintf(stderr, "  --logpcm[:fileBase]\n");
  fprintf(stderr, "    Logs UL in/out and DL in/out PCM samples.\n");
  fprintf(stderr, "    Default fileBase is pcm.\n");
  fprintf(stderr, "    PCM file name template: fileBase<src><index>.pcm\n");
  fprintf(stderr, "  --log:file\n");
  fprintf(stderr, "    Sets output log file to file.\n");
  fprintf(stderr, "    Default is results.txt.\n");
  fprintf(stderr, "    [stdout] and [stderr] can be used to send output to\n");
  fprintf(stderr, "    stdout and stderr streams.\n");
  fprintf(stderr, "  --loghlack\n");
  fprintf(stderr, "    Define to log HLACK reception result to output file.\n");
  fprintf(stderr, "    Default is FALSE.\n");
  fprintf(stderr, "  --ivspush\n");
  fprintf(stderr, "    Define to use IVS initiated signaling (push mode).\n");
  fprintf(stderr, "    Default is FALSE.\n");
  fprintf(stderr, "  --ackstop\n");
  fprintf(stderr, "    Define to stop control loop when ACK was received,\n");
  fprintf(stderr, "    which disables HLACK data reception.\n");
  fprintf(stderr, "    Default is FALSE.\n");
  fprintf(stderr, "  --startframe:n\n");
  fprintf(stderr, "    Set the number of frames to process before sending\n");
  fprintf(stderr, "    the MsgPsapTx signal to the PSAP.\n");
  fprintf(stderr, "    Default is 10 frames (200 ms).\n");
  fprintf(stderr, "  --startframehlack:n\n");
  fprintf(stderr, "    Set the number of frames to process before sending\n");
  fprintf(stderr, "    the MsgPsapHlack signal during ACK transmission.\n");
  fprintf(stderr, "    Default is 125 frames (2.5 s).\n");
  fprintf(stderr, "  --totalframes:n\n");
  fprintf(stderr, "    Set the number of frames to wait for the MsgPsapRx\n");
  fprintf(stderr, "    signal before declaring a transmission timeout.\n");
  fprintf(stderr, "    Default is 200*50 frames (200 s).\n");
  fprintf(stderr, "  --totalframesack:n\n");
  fprintf(stderr, "    Set the number of frames to process after reception\n");
  fprintf(stderr, "    of the MsgPsapRx signal.\n");
  fprintf(stderr, "    Default is 500 frames (10 s).\n");
  fprintf(stderr, "  --campaign:file\n");
  fprintf(stderr, "    Set the input campaign file to file.\n");
  fprintf(stderr, "    Default is campaign.txt\n");
  fprintf(stderr, "  --invcodec[:dir]\n");
  fprintf(stderr, "    Invert vocoder output signal (multiply by -1).\n");
  fprintf(stderr, "    Specify dir=ul or dir=dl, otherwise invert both.\n");
  fprintf(stderr, "    Default is FALSE.\n");

  va_end(args);
}

static int ProcessArgs(int argc, char *argv[]) {
  while (argc-- > 0) {
    int n;

    if (strlen(*argv) >= ARG_MAX_LENGTH) {
      Usage("Argument too long <%s>", *argv);
      return FALSE;
    }

    n = option(*argv, options, p1);

    switch (n) {
      case 0:  /* --logpcm:%s */
        strcpy(pcmFile, p1);
        pcm = pcmFile;
        break;
      case 1:  /* --logpcm */
        pcm = pcmFile;
        break;
      case 2:  /* --log:%s */
        strcpy(logFile, p1);
        break;
      case 3:  /* --loghlack */
        logHlack = TRUE;
        break;
      case 4:  /* --ivspush */
        ivsPush = TRUE;
        break;
      case 5:  /* --ackstop */
        ackStop = TRUE;
        break;
      case 6:  /* --startframe:%s */
        if (sscanf(p1, "%d", &startFrame) != 1) {
          Usage("Invalid startFrame count <%s>", p1);
          return FALSE;
        }
        break;
      case 7:  /* --startframehlack:%s */
        if (sscanf(p1, "%d", &startFrameHlack) != 1) {
          Usage("Invalid startFrameHlack count <%s>", p1);
          return FALSE;
        }
        break;
      case 8:  /* --totalframes:%s */
        if (sscanf(p1, "%d", &totalFrames) != 1) {
          Usage("Invalid totalFrames count <%s>", p1);
          return FALSE;
        }
        break;
      case 9:  /* --totalframesack:%s */
        if (sscanf(p1, "%d", &totalFramesAck) != 1) {
          Usage("Invalid totalFramesAck count <%s>", p1);
          return FALSE;
        }
        break;
      case 10: /* --campaign:%s */
        strcpy(cpnFile, p1);
        break;
      case 11: /* --invcodec:%s */
        strncpy(invCodec, p1, 4);
        break;
      case 12: /* --invcodec */
        strcpy(invCodec, "both");
        break;
      case 13: /* -h */
      case 14: /* --help */
        Usage(NULL);
        return FALSE;
      default:
        Usage("Invalid option %s", *argv);
        return FALSE;
    }
    argv++;
  }

  if (strlen(pcmFile) <= 0) strcpy(pcmFile, "pcm");
  if (strlen(logFile) <= 0) strcpy(logFile, "results.txt");
  if (strlen(cpnFile) <= 0) strcpy(cpnFile, "campaign.txt");
  if (startFrame      == 0) startFrame      = 10;
  if (startFrameHlack == 0) startFrameHlack = 125;
  if (totalFrames     == 0) totalFrames     = 10000;
  if (totalFramesAck  == 0) totalFramesAck  = 500;

  return TRUE;
}

int main(int argc, char *argv[]) {
  if (ProcessArgs(argc-1, &argv[1])) {
    LogInit(stderr);
    IpcInit();
    ControlInit("localhost", "10000", "10001", "10002", "10003",
      pcm, cpnFile, logFile, invCodec, logHlack, ivsPush, ackStop,
      totalFrames + startFrame, totalFramesAck,
      startFrame, startFrameHlack);

    UlChannelInit("localhost", "10000");
    DlChannelInit("localhost", "10001");
    IvsInit("localhost", "10002");
    PsapInit("localhost", "10003");

    IpcStart();
    ControlRun();
    IpcStop();
    UlChannelFinal();
    DlChannelFinal();
    IpcFinal();

    return 0;
  }
  return -1;
}
